//=============================================================================
// GLevelSecurity.
//	Author:	Pcube
//	Abstract:
//		This SecurityHandler object provides AuthorityTable functionality as in
//	xCoop.
//=============================================================================
class GLevelSecurity expands GSecurityHandler
	config(GLevelSecurity);
	
	// Imports
#exec obj load file="..\sharedSystem\ipPlugin.u"
	
//=============
// Constants
	// Misc
const VERSION_NUMBER = "2.22";
const NAME_SECUREBROADCAST = 'GSecurityHandler';
const NAME_COMPATBROADCAST = 'Event';

const CLASS_UTILITY = class'GUtility';
const CLASS_IPPLUGIN = Class'IPPlugin';
const CLASS_QUERYOBJECT = class'GLSUDPServerQuery';

const LOG_WATERMARK = 'GLevelSecurity';

const ERR_FAILED_TO_SPAWN_SECURITYINFO = "Failed to create GLSSecurityInfo object for player ";
const ERR_BAD_PLUGIN = "Warning; bad IPPlugin class variable. Failed to load IPPlugin!";

const MSG_SELF_DESCRIBING_PRONOUN = "You";
const MSG_INVALID_ID = "Invalid ID.";
const MSG_INSUFFICIENT_ADMIN_RIGHTS = "You are not allowed to perform that action on the target player.";
const MSG_INVALID_GAMEPASSWORD = "Incorrect password.";
const MSG_VALID_GAMEPASSWORD = "Password accepted.";
const MSG_CAPACITY = "Server at capacity.";
const MSG_NO_GAMEPASSWORD = "No game password required";
const MSG_PRELOGIN_CONNECTING = "is connecting to this server...";
const MSG_SPAM_VIOLATION = "You have been temporarily muted for spamming.";
const MSG_SPAM_UNVIOLATION = "You are now unmuted.";
const MSG_ADMIN_REVOKED = "Your administrative rights have been revoked.";

const MSG_NOT_IN_SERVERPACKAGES = "The playerclass you wish to use is not being run on the server.";
const MSG_PLAYERCLASS_NOT_ALLOWED = "The playerclass you wish to use is not allowed on this server.";

const MSG_VOTE_CAST = "voted to end the level.";
const MSG_VOTE_REVOKED = "revoked their vote to end the level.";

const MIN_TIME_BETWEEN_CONNECTIONS = 1.0;

const MAX_REPETITIONS_BEFORE_RJA_BAN = 3;

	// Admin levels
const CLIENT_ADMINLEVEL = 4;
const TERTIARY_ADMINLEVEL = 3;
const SECONDARY_ADMINLEVEL = 2;
const PRIMARY_ADMINLEVEL = 1;


//=============
// Variables

const CLASS_SECURITYINFO = class'GLevelSecurity_Shared.GLSSecurityInfo';

	// Authority table
struct S_AuthorityKey
{
	var() config int ActionType;
	var() config string Action;										// Tag of the action described by this entry
	var() config int MinAdminLevel;										// Minimum admin level required
	var() config bool bCovert;											// If true, is not broadcasted to admins
};

enum E_ActionLoggingType
{
	ALT_None,
	ALT_ValidatedOnly,
	ALT_RejectedOnly,
	ALT_All
};

struct S_AuthorityValidationPacket
{
	var bool bAccepted;
	var bool bCovert;
};

var(ActionControl) private config S_AuthorityKey AuthorityTable[256];	// Table of allowable actions & associated permission levels
var private int NumTableEntries;
var(ActionControl) private config S_AuthorityKey DefaultAction;		// If not found in table, will fall back to this default.
var(ActionControl) private config E_ActionLoggingType ActionLoggingType;
var(ActionControl) private config int MaxRejectionsPerSecond;				// If > 0, then failing a command more than X number of times per second will result in an auto kick.


struct s_IP
{
  var() config string firstsubnet,secondsubnet,thirdsubnet,fourthsubnet; // firstsubnet.secondsubnet.thirdsubnet.fourthsubnet:port
};

struct s_BanInfo // An IPv4 ban entry, consisting of four subnets and internal data about how this ban was recorded
{
  var() config s_IP IP;
  var() config string BanEntryDate; // Timestamp of when the ban was inserted
  var() config string BannedBy; // Name of the admin who banned this player
  var() config string BanReply; // What to tell this asshole when they try to connect (rejection message)
};

struct s_IPAdminInfo
{
  var() config s_IP IP;
  var() config int AdminLevel;
};

	// Controls for incoming connections and logging in as administrator
var(LoginControl) config s_BanInfo BanTable[1024];  // Up to 256 bans may be entered
var(LoginControl) config string IPPluginClass;  // IPPlugin to use with this mutator
var Class<IPPlugin> IPPlugin; // IPPlugin to use

var(LoginControl) private config string AdminPassword[3];			// 3 passwords corresponding to admin levels 3, 2, and 1.
var(LoginControl) private config string GamePassword;					// Password required to enter the game
var(LoginControl) private config bool bAutoBanRJA;
var(LoginControl) private config int MinbAdminLevel;					// Must be at least this level to make bAdmin take effect
var(LoginControl) private config int MaxPlayers;
var(LoginControl) private config int MaxSpectators;		// Unused.
var(LoginControl) private config bool bAllowOnlyGPlayerSubclasses;		// If true, only GPlayer subclasses are permitted.

var(LoginControl) private config bool bUseIPAdminTable;			// If true, "login" command will be disabled. All admins will automatically log in by lookup in the IP Admin table.
var(LoginControl) private config s_IPADminInfo IPAdminTable[32];		// Lookup table mapping IPs to AdminLevels.

	// Controls over spam and chat
var(ChatControl) private config int MaxMessagesPerSecond;				// Max # of messages allowed per sec.
var(ChatControl) private config int MinAntiSpamOverrideLevel;		// Min level to override anti-spam (set to 3 for anti-spam off)
var(ChatControl) private config int AntiSpamMuteTime;					// Time in seconds to punish a spam attempt
var(ChatControl) private config bool bChatLog;

	// Controls pertaining to how messages are sent to clients/logged.
var(ReportingControl) private config string GeneralSecureMessage;			// Preptended message to general secure command messages, i.e. ("[Secure]")
var(ReportingControl) private config bool bSilentSecureMessages;			
var(ReportingControl) private config int MinReceiveSecurityEventsLevel;			// Must be at least this level of admin to receive security events.
var(ReportingControl) private config bool bBroadcastAuthorizations;			// If true, all authorized actions will be reported to admins.
var(ReportingControl) private config bool bHUDCompatibilityMode;				// If true, all broadcasted messages are of type Event.

	// Game modifying controls
var(GameControl) private config bool bForceSeedNoRespawn;		// If true, seeds will never respawn.
var(GameControl) private config bool bVoteEnds;					// If true, players must vote to authorize ends
var private bool bVoteEnds_set;
var(GameControl) private config byte MinVotePercentage;		// % of players needed to vote to transition map

// Runtime Variables
var private GLSSecurityInfo SuperAdmin;
var private bool bAllowEnds;				// if true, ending is possible
var private int PreLoginRepetitions;

struct S_EventTimestamp
{
	var string Event;
	var float Time;
};
var S_EventTimestamp PreloginCache;

//================
// Event Handling
Function Initialize( GenericGame Game, String Options )
{
	super.initialize(game,Options);
	RehashTable();
	
	SuperAdmin = spawn(CLASS_SECURITYINFO);
	SuperAdmin.SetAdminLevel(0);
	
	bAllowEnds = true;
    IPPlugin = Class<IPPlugin>(DynamicLoadObject(IPPluginClass,class'CLASS'));
	LogMarked("Using IPPlugin class"@IpPlugin);
	if( bUseIPAdminTable )
	{
		if( IPPlugin == None )
		{
			Warn("Warning!!! IPAdminTable attempted to be used without a valid IPPlugin selected! Logging in as admin is impossible!");
		}
		else
		{
			LogMarked("Using IPAdminTable to automatically log in incoming clients. \"Login\" command disabled.");
		}
	}
	PreLoginRepetitions = 0;
	bVoteEnds_set = bVoteEnds;
	SaveConfig();
}

// GetRenderDescription
//	Called once at game startup. The string that you return with this method is
//	what rendering elements will (optionally) render to the HUD. Return the
//	empty string to not display anything about this handler (i.e. hide it from
//	being rendered). Return a short description to allow the description to be
//	viewed publically.
Function String GetRenderDescription()
{ 
	local String s;
	s = "GLevelSecurity v"$VERSION_NUMBER;
	if( bVoteEnds_set )
	{
		s = s@"(EndVote)";
	}
	return S;
}

// GetMaxPlayers
//	Called when game initially needs to have its variables updated 
//	(level.game.maxplayers).
Function int GetMaxPlayers()
{
	return MaxPlayers;
}

// GetMaxSpectators
//	Called when game initially needs to have its variables updated 
//	(level.game.maxspectators).
Function int GetMaxSpectators()
{
	return MaxSpectators;
}

// GetSecurityInfoObject
//	Returns a newly created Security Info object for a specific player, Other.
//	Useful for attaching your own security info object to playerpawns so you
//	can record metadata about each player.
Function GSecurityInfo GetSecurityInfoObject(pawn Other)
{
local GLSSecurityInfo obj;
	obj = spawn(CLASS_SECURITYINFO,Other,,Other.location,Other.rotation);
	if( obj == none )
	{
		WarnMarked(ERR_FAILED_TO_SPAWN_SECURITYINFO$Other.GetHumanName());
		return none;
	}
	obj.SetVoting( bVoteEnds_set );
	return obj;
}

// IsRelevant:
//	Return whether an actor should be destroyed in this type of game.
function bool IsRelevant( actor Other )
{
	return true;
}

// LogGameParameters:
//	Called when gamestate needs to be logged. Override this to provide custom
//	logging support when the game needs to be logged.
function LogGameParameters(StatLog StatLog);


//====================================================================================
// PLAYER ENTRY & EXIT

// PreLogin:
//	Handles preprocessing of join requests. Called when a player initially makes 
//	a connection attempt. Set Error to a non empty string to reject the login 
//	attempt. Return true to allow, false and set error to reject.
Function bool PreLogin( out String PlayerOptions, out string Error )
{
local string Pass;
local string PName;
local int index;
local string Temp;
local float DeltaTime;
local int hits;

	// Get last prelogin IP
	if( isBanned(none,Error) )
	{
	    // Log and broadcast this event
        SecurityEvent("Incoming connection from"@CLASS_UTILITY.static.parseOption(PlayerOptions,"name")@"was rejected. Reason:"@Error);
		return false;
	}
	
	if( bAutoBanRJA )
	{
		if( PreLoginCache.Event ~= IPPlugin.static.retrieveIP(level,none) )
		{
			DeltaTime = Level.TimeSeconds - PreLoginCache.Time;
			if( DeltaTime < MIN_TIME_BETWEEN_CONNECTIONS )
			{
				PreLoginRepetitions++;
				if( PreLoginRepetitions > MAX_REPETITIONS_BEFORE_RJA_BAN )
				{
					// Ban.
				  if( insertBan( None, IPPlugin.static.retrieveIP( level, none ), "You have been banned for performing an RJA attack. Please mail the packet to the administrator's home address." ) )
				  {
					// Success    
					SecurityEvent( "Automated ban entry (RJA attack) inserted against incoming player"@CLASS_UTILITY.static.ParseOption(PlayerOptions,"Name") );
				  }
				  else
				  {
					// Failure
					SecurityEvent( "Failed to insert automated ban (RJA attack) against"@CLASS_UTILITY.static.ParseOption(PlayerOptions,"Name") );
				  }
				}
			}
			else
				PreLoginRepetitions=0;
		}
		else
		{
			PreLoginRepetitions=0;
		}
	
		PreloginCache.Event = IPPlugin.static.retrieveIP(level,none);		// Get Prelogin IP
		PreLoginCache.Time = Level.TimeSeconds;
	}
		
	Pass = CLASS_UTILITY.static.ParseOption(PlayerOptions,"Password");
	PName = CLASS_UTILITY.static.ParseOption(PlayerOptions,"Name");

	LogMarked("New player prelogin request: "$PlayerOptions);
	AlertAll(PName@MSG_PRELOGIN_CONNECTING,true);
	
	if( level.game.NumPlayers == level.game.MaxPlayers )
	{
		AlertAll(PName@"failed login: "$MSG_CAPACITY,false);
		Error = MSG_CAPACITY;
		return false;
	}
	
	// Check for Password
	if( GamePassword != "" &&
		Pass != GamePassword )
	{
		AlertAll(PName@"failed login: "$MSG_INVALID_GAMEPASSWORD,false);
		Error = MSG_INVALID_GAMEPASSWORD;
		return false;
	}
	
	AlertAll(PName@"allowed login: "$MSG_NO_GAMEPASSWORD,false);
	return true;
}

// AllowLogin:
//	Handles preprocessing of login requests. Return true to allow the login to 
//	pass to GameLogic, false and set error to reject it from further processing.
//	This call happens efore mutators have had a chance to modify the login 
//	options of the incoming client, and before the GameLogic class has gotten 
//	the call.
Function bool AllowLogin( out string Portal, out string Options, 
	out string Error, out class<PlayerPawn> SpawnClass )
{
	LogMarked("New player login request: "$Portal@options@spawnclass);
	return true;
}

// AllowProcessedLogin:
//	Handles postprocessing of login requests. This call happens after mutators
//	have had a chance to modify the login options of the incoming client. This 
//	call only happens if the GameLogic class & mutators did not block the login.
Function bool AllowProcessedLogin( out PlayerPawn NewPlayer, string Portal, 
	string Options, out string Error, class<PlayerPawn> SpawnClass )
{
	if( level.game.NumPlayers == level.game.MaxPlayers )
	{
		AlertAll(CLASS_UTILITY.static.ParseOption(Options,"name")@"failed login: "$MSG_CAPACITY,false);
		Error = MSG_CAPACITY;
		return false;
	}

	
	if(!IsInServerPackages(left(string(NewPlayer.Class),instr(string(newplayer.class),"."))))
	{
		error = MSG_NOT_IN_SERVERPACKAGES;
		return false;
	}	
	
	if( bAllowOnlyGPlayerSubclasses && !NewPlayer.IsA('GPlayer') )
	{
		error = MSG_PLAYERCLASS_NOT_ALLOWED;
		return false;
	}
	
	LogMarked("Finalized login request: "$NewPlayer@Portal@Options@SpawnClass);
	return true;
}

// FinalizeLogin:
//	Allows postprocessing and finalization for a new playerpawn object. It is
//	safe to replicate to the playerpawn at this point.
Function FinalizeLogin( pawn NewPlayer )
{
	// Account for new player
	if( NewPlayer.IsA('PlayerPawn') )
	{
		level.game.numplayers++;
	}

	// Safely init. info
	GLSSecurityInfo( GPlayerReplicationInfo( NewPlayer.PlayerReplicationInfo ).GetSecurityInfo() ).Initialize();
	
	if( bUseIPAdminTable &&
		NewPlayer.IsA('PlayerPawn') )
	{
		// Try a login.
		AdminLogin_IP( NewPlayer, IPPlugin.static.retrieveIP(level,playerpawn(NewPlayer)) );
	}
}

// Logout:
//	Handles postprocessing of logout requests. Call happens after all mutators 
//	and gamelogic have processed logout.
function Logout( pawn Exiting )
{
	if( Exiting.IsA('PlayerPawn') )
	{
		level.game.numplayers--;
	}
	LogMarked("Player exiting: "$Exiting.GetHumanName());
}
//====================================================================================


//====================================================================================
// SPAM MANAGEMENT

// HandleMessage:
//	Called when a message has been sent. Useful to record metadata about a
//	user's chat history. Could be used to detect spam for example.
Function HandleMessage( Pawn Sender, Name MsgType, String Msg )
{
local GLSSecurityInfo Info;
local string SenderName;
local String Temp;

	if( Sender != None )
		SenderName = Sender.GetHumanName();
	else
		SenderName = "[Server]";

	// Chat logging
	if( bChatLog )
	{
		LogMarked("ChatLog:"@SenderName$":"@MSG@"("$MsgType$")");
	}

	if( MSGType != 'Say' && MSGType != 'TeamSay' )
		return;		
		
	// See if they can bypass antispam, also check for votes
	if( Sender != None )
	{
		Info = GLSSecurityInfo( GPlayerReplicationInfo( Sender.PlayerReplicationInfo ).GetSecurityInfo() );
		if( Info.GetAdminLevel() > MinAntispamOverrideLevel )
		{
			if( Info.AddMessage(MSG) > MaxMessagesPerSecond )
				BeginSpamMute( Sender );
		}
		
		// Check for VoteEnd flag
		if( Info != None &&
			bVoteEnds_set &&
			Sender.IsA('PlayerPawn') && 
			( MSG ~= "VoteEnd" ) )
		{
			Info.ToggleVote();
			
			if( Info.GetVoteStatus() )
				Temp = MSG_VOTE_CAST;
			else
				Temp = MSG_VOTE_REVOKED;
				
			AlertAll(Sender.GetHumanName()@Temp@"("$MinVotePercentage$"% required,"@GetVotingPercentage()$"% cast)",bSilentSecureMessages);
		}
	}
}

function BeginSpamMute( Pawn Other )
{
	// Create violation event
	ViolationEvent(Other,MSG_SPAM_VIOLATION);
	GLSSecurityInfo( GPlayerReplicationInfo( Other.PlayerReplicationInfo ).GetSecurityInfo() ).BeginSpamMute( AntiSpamMuteTime );
}

//====================================================================================

// ShouldRespawn:
//	Return whether an item should respawn. Useful to filter respawning of
//	certain items.
function bool ShouldRespawn( actor Other )
{
	if( bForceSeedNoRespawn &&
		Other.IsA('Seeds') )
		return false;
	return Super.ShouldRespawn(Other);
}

//===================
// Action Validation

// HasAuthority:
// 	Returns True if Other has authority to perform the action specified. Returns
//	false and sets Error equal to the reason why the validation failed if Other 
//	isn't allowed to perform the action.
// Parameters:
//	Other - The player that wishes to perform the action.
//	ActionType - A flag denoting the type of this action (Player, Security, or
//					Custom)
//	Action - A name corresponding to an action.
//	ActionData - Extra information relevant to the action given.
//	Error - A reason why authentication failed (if it failed)
Function bool HasAuthority( pawn Other, int ActionType, string Action, 
	optional coerce string ActionData, optional out string Error )
{
local string status;
local string extra;
local string username;
local string type;
local S_AuthorityValidationPacket Packet;
local bool result;

	// Quick check to stop spammers.
	if( Other != None )
	{
		userName = Other.GetHumanName();
		if( Other.bDeleteMe )
			return false;
	}
	else
		UserName = "Server";
		
	Packet = Wrapped_HasAuthority(Other,ActionType,Action,ActionData,Error);
	result = Packet.bAccepted;
	// Perform logging.
	if(result)	status="Accepted";
	else		status="Rejected";
	
	
	if( !Packet.bCovert &&
		bBroadcastAuthorizations &&
		!(Action~="PSayID") )
	{
		// Broadcast all authority events to admins
		switch(ActionType)
		{
			case(ACTIONTYPE_SECURE):
				type = "secure";
				break;

			case(ACTIONTYPE_MUTATE):
				type = "mutator";
				break;
				
			case(ACTIONTYPE_CUSTOM):
				type = "custom";
				break;

			case(ACTIONTYPE_PLAYER):
				type = "player";
				break;			

			default:
				type = "";
				break;				
		}
		
		SecurityEvent(username@"requested authority for"@type@"action"@Action@ActionData@"("$status$")");
	}
	
	// Logging.
		// I _really_ hate hardcoding things, but privacy is important. Fuck server logging of psays.
		// I take no responsibility for any admin's choice to remove this hardcoding, but I strongly disapprove of it.
	if( !(Action ~= "PSayID") && (
		ActionLoggingType == ALT_All ||
		(ActionLoggingType == ALT_ValidatedOnly && result) ||
		(ActionLoggingType == ALT_RejectedOnly && !result)
		) )
	{
		if( Action != "" )
			Action = Action$" ";
		if( ActionData != "" )
			ActionData = ActionData$" ";
		LogMarked(status$":"@Other.GetHumanName()@"performing"@Action$ActionData$"of type"@string(ActionType));
	}
	
	if( !Result )
	{
		AlertClient(Other,"Command rejected:"@Action@ActionData,bSilentSecureMessages);
		if( MaxRejectionsPerSecond > 0 && !Other.bDeleteMe )
			if( GLSSecurityInfo( GPlayerReplicationInfo(Other.PlayerReplicationInfo).GetSecurityInfo() ).AddCommand(Action,ActionData,Result) > MaxRejectionsPerSecond )
			{
				// Auto-kick!
				ViolationEvent(Other,"You were automatically kicked for spamming commands.");
				Other.Destroy();
				return false;
			}
	}
			
	return result;
}

function S_AuthorityValidationPacket Wrapped_HasAuthority( pawn Other, int ActionType, String Action, 
	optional coerce string ActionData, optional out string Error )
{
local int pivot;
local int UpperBound,LowerBound;
local int result;
local S_AuthorityKey Key;
local bool bFound;
local S_AuthorityValidationPacket ret;

	// Binary search into the table
	UpperBound = NumTableEntries;
	LowerBound = 0;
	
	while((UpperBound - LowerBound) > 0)
	{
		Pivot = LowerBound + ((UpperBound - LowerBound) / 2);
		Result = CLASS_UTILITY.static.strcmp( caps(Action$ActionType), caps(AuthorityTable[Pivot].Action$AuthorityTable[Pivot].ActionType) );
		
		if( Result == 0 )	// Hit!
		{
			bFound = true;
			Key = AuthorityTable[Pivot];
			break;
		}
		else if( Result < 0 )
		{
			// It's less.
			if( UpperBound - LowerBound == 1 )
				break;
			UpperBound = Pivot;
		}
		else
		{
			// It's greater.
			if( UpperBound - LowerBound == 1 )
				break;
			LowerBound = Pivot;
		}
	}

	// Fuck oldunreal
	if( !bFound )
		Key = DefaultAction;
	
	// Validate
	ret.bAccepted = ( GLSSecurityInfo( GPlayerReplicationInfo( Other.PlayerReplicationInfo ).GetSecurityInfo() ).GetAdminLevel() <= Key.MinAdminLevel ) &&
					( Key.ActionType == ACTIONTYPE_UNKNOWN ||
						Key.ActionType == ActionType );
	ret.bCovert = key.bCovert;
	return ret;
}

// AllowsBroadcast:
//	Called when a message wishes to be broadcasted. Return false to reject.
//	Useful for blocking unwanted messages.
function bool AllowsBroadcast( Actor Broadcaster, int Len )
{
	// Broadcast filtering - Disallow player owned objects.
	if ( len==0 || (!Broadcaster.IsA('PlayerPawn') && Broadcaster.Owner != None && Broadcaster.Owner.IsA('PlayerPawn')) )
		return false;

	if( Pawn(Broadcaster) != None &&
		GPlayerReplicationInfo( Pawn(Broadcaster).PlayerReplicationInfo ).GetSecurityInfo() != none && 
		GLSSecurityInfo( GPlayerReplicationInfo( Pawn(Broadcaster).PlayerReplicationInfo ).GetSecurityInfo() ).Muted() &&
		GLSSecurityInfo( GPlayerReplicationInfo( Pawn(Broadcaster).PlayerReplicationInfo ).GetSecurityInfo() ).GetAdminLevel() > MinAntispamOverrideLevel )
		return false;
	return true;
}

// AllowChangeTeam:
//	Return whether a team change is allowed. Return true to allow, false to 
//	deny.
function bool AllowChangeTeam(Pawn Other, int N)
{
	return true;
}

// AllowRestartGame:
//	Return true to allow Gamelogic to restart the game, false to reject.
function bool AllowRestartGame()
{
	return true;
}

// AllowLevelSwitch:
//	Return true to allow a level to switch. False to reject. Called before 
//	mutators or gamelogic can act on request.
function bool AllowLevelSwitch( pawn Other, String URL, 
	bool bTeleporterEnd )
{
local GLSSecurityInfo Info;

	if( !bTeleporterEnd )
		SecurityEvent(Other.GetHumanName()@"attempting to switch level by command to"@URL);
	else
	{
		SecurityEvent(Other.GetHumanName()@"attempting to switch level manually to"@URL);
		AlertAll(Other.GetHumanName()@"attempting to switch level manually to"@URL,bSilentSecureMessages);
	}

	if( !IsValidURL( URL ) )
	{
		SecurityEvent("Blocking level switch to URL. Invalid URL specified.");
		return false;
	}
	
	if( bTeleporterEnd )
	{
		if( !bAllowEnds )
		{
			AlertClient(Other,"Level end attempt rejected: The ending is disabled.",bSilentSecureMessages);
			return false;
		}
	
		if( bVoteEnds_set )
		{
			Info = GLSSecurityInfo( GPlayerReplicationInfo( Other.PlayerReplicationInfo ).GetSecurityInfo() );
			if( Info.GetVoteStatus() == False )
			{
				Info.ToggleVote();
				AlertAll(Other.GetHumanName()@"voted to end the level by touching the end.",bSilentSecureMessages);
			}
			
			if( GetVotingPercentage() < MinVotePercentage )
			{	
				AlertAll("Level end attempt rejected: Not enough clients have voted to end the level. ("$MinVotePercentage$"% required,"@GetVotingPercentage()$"% cast)",bSilentSecureMessages);
				AlertAll("Say \"VoteEnd\" in the chat to cast your vote.",bSilentSecureMessages);
			
				return false;
			}
		}
	}
	
	return true;
}

Function byte GetVotingPercentage()
{
local playerpawn P;
local GLSSecurityInfo Info;
local int NumVotes,NumTotal;
	NumVotes = 0;
	NumTotal = 0;
	foreach allactors(class'PlayerPawn',P)
	{
		Info = GLSSecurityInfo( GPlayerReplicationInfo(P.PlayerReplicationInfo).GetSecurityInfo() );
		if( Info.GetVoteStatus() )
			NumVotes++;
		NumTotal++;
	}
	
	return byte( 100.*(float(NumVotes)/float(NumTotal)) );
}

// Returns whether the proposed map to switch to is indeed valid
function bool isValidURL( string URL )
{
local int index;

  // Parse appropriate map name.
  index = instr(URL,"#");
  if( index != -1 )
    url = left(url,index);
    
  index = instr(URL,"?");
  if( index != -1 )
    url = left(url,index);
    
  index = instr(URL,"/");
  if( index != -1 )
    url = left(url,index);

  // Attempt load
  return (dynamicLoadObject(Trim(URL)$".MyLevel",class'level') != none);
}

static final function string LTrim(coerce string S)
{
	return CLASS_UTILITY.static.LTrim(S);
}

static final function string RTrim(coerce string S)
{
	return CLASS_UTILITY.static.RTrim(S);
}

static final function string Trim(coerce string S)
{
	return CLASS_UTILITY.static.Trim(S);
}

//======================================================================================
// Commands

// HandleSecureCommand:
//	Allows handling of base secure commands. Useful to implement your own
//	security related commands, such as kick, kill, ban, etc...
Function HandleSecureCommand( pawn Other, string Command, 
	optional coerce string Options )
{
	if( GLSSecurityInfo(GPlayerReplicationInfo(Other.PlayerReplicationInfo).GetSecurityInfo()) == None )
	{
		WarnMarked(Other@"does not have a SecurityInfo object!");
		return;
	}

	Wrapped_HandleSecureCommand( GLSSecurityInfo(GPlayerReplicationInfo(Other.PlayerReplicationInfo).GetSecurityInfo()), Command, Options );
}

Function Wrapped_HandleSecureCommand( GLSSecurityInfo User, string Command, optional coerce string Options )
{
local string newname;
local string Pass;
local string temp;
local pawn Target;
local pawn Other;
local string SenderName;
local GLSSecurityInfo TargetInfo;

	if( User == None )
	{
		WarnMarked("Wrapped_HandleSecureCommand called with null User SecurityInfo!");
		return;
	}
	
	if( User != SuperAdmin )
	{
		Other = pawn(User.Owner);
		SenderName = Other.GetHumanName();
	}
	else
	{
		Other = None;
		SenderName = "Server";
	}

	if( Options != "" )
		temp = " "$Options;
	else
		temp = "";

	// Parse ID		
	if( Right(Command,len("ID")) ~= "ID" )
	{
		temp = Options;
		if( instr(temp," ") != -1 )
			temp = left(temp,instr(temp," "));
		if( temp == "" )
			goto 'Invalid';
		Target = GenericGame(Level.Game).GetPawnFromID(int(temp));
		if( Target == None )
		{
			Invalid:
			AlertClient(Other,MSG_INVALID_ID,bSilentSecureMessages);
			return;		
		}
		
		TargetInfo = GLSSecurityInfo(GPlayerReplicationInfo(Target.PlayerReplicationInfo).GetSecurityInfo());
		
		if( !User.HasAuthorityOver(TargetInfo) && caps(command) != "PSAYID" )
		{
			AlertClient(Other,MSG_INSUFFICIENT_ADMIN_RIGHTS,bSilentSecureMessages);
			return;
		}
	}
	
	switch(CAPS(Command))
	{
		case("LOGIN"):
			if( !bUseIPAdminTable )	// Disable on lookup scheme.
				AdminLogin_Pass(Other,Options);
			break;
		
		case("LOGOUT"):
			AdminLogout(Other);
			break;
			
		case("ADMIN"):
			Admin(Other,Options);
			break;
			
		case("GETIDS"):
			DisplayIDs(Other);
			break;
		
		case("GETIPS"):
			DisplayIPs(Other);
			break;
		
		case("LISTBANS"):
			ListBans(Other);
			break;
			
		case("BANID"):
			BanID(Other,Target,mid(options,1+instr(options," ")));
			break;
			
		case("BANIP"):
			BanIP(Other,Options);
			break;
			
		case("REMOVEBAN"):
			DeleteBan(Other,Options);
			break;
			
		case("END"):
			End(Other,Options);
			break;		
		
		case("ASAY"):
			ASay(Other,mid(options,1+instr(options," ")));
			break;
			
		case("PSAYID"):
			PSayID(Other,Target,mid(options,1+instr(options," ")));
			break;
		
		case("KILLALL"):
			KillAll(Other,options);
			break;
			
		case("SUMMONID"):
			SummonID(Other,Target);
			break;

		case("GOTOID"):
			GoToID(Other,Target);
			break;
			
		case("RESTARTID"):
			RestartID(Other,Target);
			break;
			
		case("GIVEADMINID"):
			temp = Options;
			temp = Mid(temp,instr(temp," ")+1);
			GiveAdminID(Other,Target,int(temp));
			break;
			
		case("REVOKEADMINID"):
			RevokeAdminID(Other,Target);
			break;
			
		case("KILLID"):
			KillID( Other, Target );
			break;
			
		case("KICKID"):
			KickID( Other, Target );
			break;
			
		case("MUTEID"):
			MuteID( Other, Target );
			break;
			
		case("UNMUTEID"):
			UnmuteID( Other, Target );
			break;
			
		case("FREEZEID"):
			FreezeID( Other, Target );
			break;
			
		case("UNFREEZEID"):
			UnfreezeID( Other, Target );
			break;
			
		case("RENAMEID"):
			newname = mid(options,1+instr(options," "));
			RenameID( Other, Target, NewName );		
			break;
			
		case("SETNAME"):
			AlertAll(Other.GetHumanName()@"has changed name to"@Options,false);
			Other.PlayerReplicationInfo.PlayerName = Options;
			if( Other.IsA('PlayerPawn') )
				GPlayerReplicationInfo(Other.Playerreplicationinfo).GetFrameworkSupportObject().RemoteUpdateURL("Name",Options,true);
			break;
			
		case("HELP"):
			DisplayCommands(Other,Options);
			break;
	}
}

final function DisplayCommands( pawn Requester, string Specific )
{
local string buf,next;
local int i;
local int min;
	buf="";
	if( Requester == None ) min = 0;
	else min = GLSSecurityInfo(GPlayerReplicationInfo(Requester.PlayerReplicationInfo).GetSecurityInfo()).GetAdminLevel();
	AlertClient(Requester,"-====GLevelSecurity Help====-",bSilentSecureMessages);
	AlertClient(Requester,"The following is a list of commands that you can use at your current admin level ("$min$")",true);
	AlertClient(Requester,"Commands with an asterisk (*) are covert and command authorization will not be displayed to admins.",true);
	AlertClient(Requester,"(PSays are never logged or displayed to admins)",true);
	AlertClient(Requester,"",true);
	for(i=0;i<NumTableEntries;i++)
	{
		if( AuthorityTable[i].MinAdminLevel >= min )
		{
			next = AuthorityTable[i].Action;
			switch(AuthorityTable[i].ActionType)
			{
				case ACTIONTYPE_SECURE:
					next = "secure"@next;
					break;
					
				case ACTIONTYPE_CUSTOM:
					next = "custom"@next;
					break;
					
				case ACTIONTYPE_PLAYER:
					break;
					
				case ACTIONTYPE_MUTATE:
					next = "mutate"@next;
					break;
					
				case ACTIONTYPE_UNKNOWN:
					break;
			}
			
			if( AuthorityTable[i].bCovert )
				next = next$"*";
			
			if( len(buf) + len(next) > 64 )
			{
				// output buf
				AlertClient(Requester,Buf,True);
				buf = next;
			}
			else if( buf != "" ) buf = buf$", "$next;
			else buf = next;
		}
	}
	if( Buf != "" )
		AlertClient(Requester,Buf,true);
	AlertClient(Requester,"-====================-",true);
}

final function DisplayIDs( pawn Requester )
{
local GPlayerReplicationInfo GPRI;
local SetIterator Iter;
	Iter = GenericGame(Level.Game).GetListReplicationInfo().GetPlayerReplicationInfoIterator();
	AlertClient(Requester,"== ID Table ==",bSilentSecureMessages);
	while( Iter.HasNext() )
	{
		GPRI = GPlayerReplicationInfo(Iter.Next());
		AlertClient(Requester,GPRI.PlayerName$":"@GPRI.GetID(),true);
	}
	GenericGame(Level.Game).GetListReplicationInfo().FreeIterator(Iter);
}

/* displayIPs
 * Prints a list of all currently connected players' IPs to this administrator
 */
final function DisplayIPs( pawn Requester )
{
local pawn pending; // For iteration

  AlertClient(Requester,"== IP Table ==",bSilentSecureMessages);

  for( pending = level.pawnList; pending != none; pending = pending.nextPawn )
  {
    if( pending.isA('PlayerPawn') )
      AlertClient(Requester,pending.getHumanName()$" - "$IPPlugin.static.retrieveIP(level,playerpawn(pending)),true);
  }
}

/* listBans
 * Prints a list of all bans on administrator's console
 */
function listBans( pawn administrator )
{
local int index; // for array traversal

	AlertClient(Administrator,"== Ban Table ==",bSilentSecureMessages);
  for( index = 0; index < arraycount( bantable ); index++ )
  {
    if( banTable[index].banEntryDate != "" )
    {
      AlertClient(Administrator,"Ban #"$index@banTable[index].IP.firstSubnet$"."$banTable[index].IP.secondSubnet$"."$banTable[index].IP.thirdSubnet$"."$banTable[index].IP.fourthSubnet,true);
      AlertClient(Administrator,"Banned on "$banTable[index].banEntryDate$" by "$banTable[index].bannedBy@"with message"@banTable[index].banReply,true);
    }
  }
}

final function End( pawn Sender, String c )
{
local string s;
	if( Sender == None ) s = "Server";
	else s = Sender.GetHumanName();
	if( c ~= "On" )
	{
		SecurityEvent(S@"turned the ends on");
		AlertAll("Ends are now enabled.",bSilentSecureMessages);
		bAllowEnds = true;
	}
	else
	{
		SecurityEvent(S@"turned the ends off");
		AlertAll("Ends are now disabled.",bSilentSecureMessages);
		bAllowEnds = false;
	}
}

final function bool AdminLogin_Pass( pawn Other, coerce string Pass )
{
local int i;
	for( i = 0; i < arraycount(AdminPassword); i++ )
	{
		if( Pass == AdminPassword[i] )
		{
			// They win
			break;
		}
	}
	i++;
	
	AdminLogin( Other, i );
}

final function bool AdminLogin_IP( pawn Other, String IPResult )
{
local s_IP parsedIP; // Four separated subnets in one structure.. parsed version of IPResult
local int index;

  parsedIP.firstSubnet = CLASS_UTILITY.static.parseNextWord( IPResult, "." );
  parsedIP.secondSubnet = CLASS_UTILITY.static.parseNextWord( IPResult, "." ); // Get second subnet
  parsedIP.thirdSubnet = CLASS_UTILITY.static.parseNextWord( IPResult, "." ); // Get first subnet
  parsedIP.fourthSubnet= CLASS_UTILITY.static.parseNextWord( IPResult, ":" );  // get fourth subnet

  for(index=0;index<arraycount(IPAdminTable);index++)
  {
    if( IPAdminTable[index].AdminLevel > 0 &&
		IPAdminTable[Index].AdminLevel < 4 )	// check  valid entry
    {
      // Check against this entry
      if( int(IPAdminTable[index].IP.firstSubnet) == int(parsedIP.firstSubnet) &&
          int(IPAdminTable[index].IP.secondSubnet) == int(parsedIP.secondSubnet) &&
          int(IPAdminTable[index].IP.thirdSubnet) == int(parsedIP.thirdSubnet) &&
          int(IPAdminTable[index].IP.fourthSubnet) == int(parsedIP.fourthSubnet) )
      {
		// Login!
		AdminLogin( Other, IPAdminTable[index].adminlevel );
        return true;
      }
    }
  }
  
  return false;
}

final function bool AdminLogin( pawn Other, int NewLevel )
{
	// successful login?
	if( NewLevel == CLIENT_ADMINLEVEL )
	{
		// Failed login
		SecurityEvent(Other.GetHumanName()@"failed to log in as an administrator");
		AlertClient(Other,"You have failed to log in as an administrator.",bSilentSecureMessages);
		return false;
	}
	else
	{
		SecurityEvent( Other.GetHumanName()@"logged in as an administrator (l"$NewLevel$")" );
		AlertClient(Other,"You have successfully logged in as a level "$NewLevel$" administrator.",bSilentSecureMessages);
		GLSSecurityInfo(GPlayerReplicationInfo(Other.PlayerReplicationInfo).GetSecurityInfo()).SetAdminLevel(NewLevel);
		if( playerpawn(other) != none )
		{
			if( NewLevel < MinbAdminLevel )
			{
				SecurityEvent("Assigning bAdmin to player"@Other.GetHumanName()@"as per MinbAdminLevel setting.");
				playerpawn(Other).bAdmin = true;
			}
			else
			{
				playerpawn(Other).bAdmin = false;
			}
		}
		return true;
	}
}

final function AdminLogout(pawn Other)
{
	if( GLSSecurityInfo(GPlayerReplicationInfo(Other.PlayerReplicationInfo).GetSecurityInfo()).GetAdminLevel() < CLIENT_ADMINLEVEL )
	{
		SecurityEvent(Other.GetHumanName()@"logged out from admin");
		GLSSecurityInfo(GPlayerReplicationInfo(Other.PlayerReplicationInfo).GetSecurityInfo()).SetAdminLevel(CLIENT_ADMINLEVEL);
		if( other.isa('playerpawn') )
			playerpawn(Other).bAdmin = false;
	}
	AlertClient(Other,"You have successfully logged out from being an administrator.",bSilentSecureMessages);
}

final function Admin(pawn Other, String Command)
{
local string Result;
	Result=Other.ConsoleCommand(Command);
	if( Result != "" )	
		AlertClient(Other,Result,bSilentSecureMessages);
}

final function GiveAdminID( pawn Sender, pawn Target, int AdminLevel )
{
local string S;
	if( Sender == None ) s = "Server";
	else s = Sender.GetHumanName();
	if( Target != None &&
		( (Sender==None || AdminLevel > GLSSecurityInfo(GPlayerReplicationInfo(Sender.PlayerReplicationInfo).GetSecurityInfo()).GetAdminLevel()) &&
		  (AdminLevel < GLSSecurityInfo(GPlayerReplicationInfo(Target.PlayerReplicationInfo).GetSecurityInfo()).GetAdminLevel()) ) 
	)
	{
		SecurityEvent(S@"gave"@Target.GetHumanName()@"level"@AdminLevel@"admin.");
		AlertClient(Target,"You have been granted level"@AdminLevel@"admin.",bSilentSecureMessages);
		GLSSecurityInfo(GPlayerReplicationInfo(Target.PlayerReplicationInfo).GetSecurityInfo()).SetAdminLevel(AdminLevel);
	}
}

final function RevokeAdminID( pawn Sender, pawn Target )
{
local string S;
	if( Sender == None ) s = "Server";
	else s = Sender.GetHumanName();
	if( Target != None )
	{
		SecurityEvent(S@"revoked"@Target.GetHumanName()@"admin rights.");
		AlertClient(Target,MSG_ADMIN_REVOKED,bSilentSecureMessages);
		GLSSecurityInfo(GPlayerReplicationInfo(Target.PlayerReplicationInfo).GetSecurityInfo()).SetAdminLevel(CLIENT_ADMINLEVEL);
	}
}

final function PSayID(pawn Sender, Pawn Target, string Message)
{
local string S,t;
local string TotalMessage;
	if( Sender == None ) s = "Server";
	else s = Sender.GetHumanName();
	if( target == None ) t = "Server";
	else t = target.GetHumanName();
	TotalMessage=S$":"@Message;
	if( Target != None )
		AlertClient(Target,"PSay from "$TotalMessage,bSilentSecureMessages);
	else
		AlertClient(None,"PSay from "$S$":"@Message,bSilentSecureMessages);
	AlertClient(Sender,"PSay to "$T$":"@Message,bSilentSecureMessages);
}

final function ASay(pawn Sender, string Message)
{
local string S;
local string TotalMessage;
	if( Sender == None ) s = "Server";
	else s = Sender.GetHumanName();
	TotalMessage="AdminChat"@s$":"@Message;
	AlertGroup(TotalMessage,TERTIARY_ADMINLEVEL,bSilentSecureMessages);
	AlertClient(None,TotalMessage,bSilentSecureMessages);
}

final function KillAll(pawn Other,String options)
{
local class<Actor> AClass;
local Actor A;
local bool bShorthandWorked;
local string s;
	if( other == None ) s = "Server";
	else s = other.GetHumanName();
	SecurityEvent(s@"killed all"@options);
	bShorthandWorked=false;
	ForEach AllActors(class	'Actor', A)
		if ( Mid( string(a.class), InStr(string(A.Class),".")+1 ) ~= Options )
		{
			bShorthandWorked=true;
			A.Destroy();
		}
	if( !bShorthandWorked )
	{
		// Try the old method
		aClass = none;
		if( InStr(Options,".") > -1 )
			AClass = class<Actor>(DynamicLoadObject(Options,class'Class'));
		if( AClass==None )
			AClass = class<Actor>(DynamicLoadObject("UnrealShare."$Options,class'Class'));
		if( AClass==None )
			AClass = class<Actor>(DynamicLoadObject("UnrealI."$Options,class'Class'));
		if( AClass==None )
			AClass = class<Actor>(DynamicLoadObject("Engine."$Options,class'Class'));
		if( AClass == None )
			return;
		ForEach AllActors(class 'Actor', A)
			if ( ClassIsChildOf(A.class, aClass) )
				A.Destroy();		
	}
}

final function SummonID(pawn Sender, pawn Target)
{
local vector Dest;
	if( Sender == None ) return;
	SecurityEvent(Sender.GetHumanName()@"summoned"@Target.GetHumanName());
	// Place Target in front of Sender.
	Dest = Sender.Location + Vector(Sender.ViewRotation)*( 
					sqrt(Sender.CollisionRadius*Sender.CollisionRadius + Sender.CollisionHeight*Sender.CollisionHeight)+
					sqrt(Target.CollisionRadius*Target.CollisionRadius + Target.CollisionHeight*Target.CollisionHeight) )*1.1;
	Target.SetLocation(Dest);
}

final function GoToID(pawn Sender, pawn Target)
{
	If( Target==None ) return;
	SecurityEvent(Sender.GetHumanName()@"teleported to"@Target.GetHumanName());
	Spawn(class'ReSpawn',,,Target.Location - (Vector(Target.Viewrotation)*2*(Sender.collisionradius+Target.collisionradius)));
	Sender.SetLocation( Target.Location - (Vector(Target.Viewrotation)*2*(Sender.collisionradius+Target.collisionradius)));
	Sender.setRotation( Rotator( Target.Location-Sender.Location ) );
}

final function RestartID(pawn Sender, pawn Target)
{
local string s;
	if( Sender == None ) s = "Server";
	else s = Sender.GetHumanName();
	SecurityEvent(S@"restarted"@Target.GetHumanName());
	Level.Game.RestartPlayer(Target);
}

final function KillID( pawn Sender, Pawn Target )
{
local string S;
	if( Sender == None ) s = "Server";
	else s = Sender.GetHumanName();
	SecurityEvent(S@"killed"@Target.GetHumanName());
	Target.Died(Sender,'',Target.location);
}

final function KickID( pawn Sender, Pawn Target )
{
local string S;
	if( Sender == None ) s = "Server";
	else s = Sender.GetHumanName();
	SecurityEvent(S@"kicked"@Target.GetHumanName());
	Target.Destroy();
}

/* banID
 * Bans a player currently in the server. Also kicks them out after the ban is complete.
 */
function bool BanID( pawn administrator, pawn Banned, optional string BanMessage )
{
local string firstChar; // The first character in the string
local string IP; // IP of the offending player
local int banID; // ID of the offender
local string s;
	if( Administrator == None ) s = "Server";
	else s = Administrator.GetHumanName();

  // See if we have the ability to even get IP info
  if( IPPlugin == None )
  {
    SecurityEvent(s@"tried to issue a ban, but there is no valid IPPlugin!");
    return false;
  }
  
  if( banned == none || 
	!banned.isa('playerpawn') )
  {
	SecurityEvent(s@"attempted to ban"@Banned.GetHumanName());
    AlertClient(Administrator,"BanID command failed. You can only ban players.",bSilentSecureMessages);
    return false;
  }
  
  // Get their info
  IP = IPPlugin.static.retrieveIP( level, playerpawn(banned) );
  
  // LET THE MIGHTY BANHAMMER FALL!!!
  if( insertBan( administrator, IP, BanMessage ) )
  {
    // Success    
    banned.destroy(); // Kick them out
    SecurityEvent( s@"issued a ban against"@banned.getHumanName() );
  }
  else
  {
    // Failure
    AlertClient(Administrator,"BanID failed. Parsing error occurred when evaluating player's IP.",bSilentSecureMessages);
  }
}

/* banIP
 * Bans a specific IP or range by directly attempting a ban insert through the insertBan method.
 */
function bool banIP( pawn administrator, string options )
{
local string s;
	if( Administrator == None ) s = "Server";
	else s = Administrator.GetHumanName();
  if( insertBan( administrator, options ) )
  {
	SecurityEvent( s@"inserted an IP ban, IP and reason:"@options );
  }
  else
  {
    SecurityEvent( s@"attempted to issue an IP ban, IP and reason:"@options );
    AlertClient( Administrator,"BanIP failed. Use BanIP #.#.#.# where each # is a subnet. The wildcard * is also usable in place of individual subnets.",bSilentSecureMessages);
  }
}

final function MuteID( pawn Sender, pawn Target )
{
local string S;
	if( Sender == None ) s = "Server";
	else s = Sender.GetHumanName();
	SecurityEvent(S@"muted"@Target.GetHumanName());
	GLSSecurityInfo(GPlayerReplicationInfo(Target.PlayerReplicationInfo).GetSecurityInfo()).Mute();
}

final function UnmuteID( pawn Sender, pawn Target )
{
local string S;
	if( Sender == None ) s = "Server";
	else s = Sender.GetHumanName();
	SecurityEvent(S@"unmuted"@Target.GetHumanName());
	GLSSecurityInfo(GPlayerReplicationInfo(Target.PlayerReplicationInfo).GetSecurityInfo()).UnMute();
}

final function FreezeID( pawn Sender, pawn Target )
{
local string S;
	if( Sender == None ) s = "Server";
	else s = Sender.GetHumanName();
	SecurityEvent(S@"froze"@Target.GetHumanName());
	Target.bMovable = false;
}

final function UnfreezeID( pawn Sender, pawn Target )
{
local string S;
	if( Sender == None ) s = "Server";
	else s = Sender.GetHumanName();
	SecurityEvent(S@"unfroze"@Target.GetHumanName());
	Target.bMovable = true;
}

final function RenameID( pawn Sender, pawn Target, string newName )
{
local string S;
	if( Sender == None ) s = "Server";
	else s = Sender.GetHumanName();
	SecurityEvent(S@"renamed"@Target.GetHumanName()@"to"@newName);
	AlertAll(Target.GetHumanName()@"has changed name to"@newName,false);
	if( Target.PlayerReplicationInfo != None )
		Target.PlayerReplicationInfo.PlayerName = NewName;
	if( Target.IsA('PlayerPawn') )
		GPlayerReplicationInfo(Target.PlayerReplicationInfo).GetFrameworkSupportObject().RemoteUpdateURL("Name",NewName,True);
}

/* deleteBan
 * Called by players. Request to remove a certain ban from the ban table.
 */
function DeleteBan( pawn administrator, string options )
{
local int slot; // proposed slot to remove the ban from
local string s;
	if( Administrator == None ) s = "Server";
	else s = Administrator.GetHumanName();

  if( !CLASS_UTILITY.static.isNumber( left( options, 1 ) ) )
  {
    AlertClient(Administrator,"Invalid ban table slot. Use DeleteBan <int slot>",bSilentSecureMessages);
    return;
  } 
  
  slot = int(CLASS_UTILITY.static.parseNextWord(options));
  
  if( removeBan( administrator, slot ) )// Perform removal
  {
    SecurityEvent( s@"removed a ban from slot"@slot );
  }
  else
  {
    SecurityEvent( s@"attempted to removed a ban from slot"@slot );
  }
}

//=================================================================================================
// Utility


//================================================================
// Client interaction methods
	// Alert a single client to a message.
	// Alerting "none" will alert the log.
final function AlertClient( pawn Client, string Message, bool bSilent )
{
local name securitymessagetype;
	if( !bHUDCompatibilityMode )
		SecurityMessageType = NAME_SECUREBROADCAST;
	else
		SecurityMessageType = NAME_COMPATBROADCAST;

	if( PlayerPawn(Client) != None )
		PlayerPawn(Client).ClientMessage(GeneralSecureMessage@Message,SecurityMessageType,!bSilent);
	else if( Client == None )
		LogMarked(GeneralSecureMessage@Message);
}

	// Alert a specific group of clients to a message, based on admin level.
final function AlertGroup( string Message, int RequiredAdminLevel, bool bSilent)
{
local pawn p;
	for( p = level.pawnlist; p != none; p=p.nextpawn )
	{
		if( GPlayerReplicationInfo(P.PlayerReplicationInfo) != None &&
			GLSSecurityInfo( GPlayerReplicationInfo(P.PlayerReplicationInfo).GetSecurityInfo() ) != None &&
			GLSSecurityInfo( GPlayerReplicationInfo(P.PlayerReplicationInfo).GetSecurityInfo() ).GetAdminLevel() <= RequiredAdminLevel )
		{
			// This admin should be alerted.
			AlertClient( P, Message, bSilent );
		}
	}
}

	// Alerts all clients to a message.
final function AlertAll( string Message, bool bSilent )
{
	AlertGroup(Message,CLIENT_ADMINLEVEL,bSilent);
	AlertClient(None,Message,bSilentSecureMessages);
}

	// Alerts admins to a message.
final function SecurityEvent( string Message )
{
	AlertClient(None,Message,bSilentSecureMessages);		// Note - this actually alerts the log
	AlertGroup(Message,MinReceiveSecurityEventsLevel,bSilentSecureMessages);
}

	// Issues a general security event message to all admins & the player specific to the action.
final function ViolationEvent( Pawn Violator,
								string Message )
{
	Message = "Violation by "$Violator.GetHumanName()$": ("$Message$")";
	SecurityEvent(Message);
	AlertClient(Violator,Message,false);
}


//================================================================
// Internal util.

// Server
//	Called when the server owner executes a 'Server' command. Similar to a
//	player executing a command, but allows for customizable server commands.
Function Server( String Command, string Options )
{
local string Temp,Temp2,Temp3,Temp4;
local S_AuthorityKey NewKey;

	if( Command ~= "SwitchLevel" )
	{
		GenericGame(Level.Game).SwitchLevel(None,Options,False);
	}
	else if( Command ~= "AddEntry" )
	{
		Temp = Left(Options,InStr(Options," "));		// Action name
		Temp2 = Mid(Options,InStr(Options," ")+1);	// Min level
		Temp3 = Mid(Temp2,InStr(Temp2," ")+1);
		Temp3 = Mid(Temp2,InStr(Temp3," ")+1);	// bCovert
		if( Temp3 == "" )
			Temp3 = "-1";
		if( Temp4 == "" )
			Temp4 = "false";

		if( Temp2 == "" || int(Temp2) <= 0 || int(Temp2) > 4 )
		{
			LogMarked("Invalid minimum level requirement. [1-4 inclusive]");
			return;
		}
		
		// Insert new key and rehash
		NewKey.Action = Temp;
		NewKey.MinAdminLevel = int(Temp2);
		NewKey.ActionType = int(Temp3);
		if( Temp4 ~= "true" )
			NewKey.bCovert = true;
		else
			NewKey.bCovert = false;
		AuthorityTable[NumTableEntries] = NewKey;
		RehashTable();
		SaveConfig();
		LogMarked("Successfully added entry"@Temp@"min lv."@int(Temp2)@"type"@int(Temp3));
	}
	else if( Command ~= "RehashTable" )
	{
		RehashTable();
		AlertClient(None,"Rehashed authority table.",bSilentSecureMessages);
		SaveConfig();
	}
	else if( Command ~= "ClearTable" )
	{
		ClearTable();
		AlertClient(None,"Authority table cleared.",bSilentSecureMessages);
		SaveConfig();
	}
	else if( Command ~= "Lockdown" )
	{
		Lockdown();
		AlertClient(None,"Server entering lockdown!!!",bSilentSecureMessages);
		SaveConfig();
	}
	else if( Command ~= "Help" )
	{
		AlertClient(None,"-====================-",bSilentSecureMessages);
		AlertClient(None,"Server only commands:",bSilentSecureMessages);
		AlertClient(None,"AddEntry<EntryName><MinLevel><ActionType><bCovert>, RehashTable, ClearTable, Lockdown, SwitchLevel",bSilentSecureMessages);
		AlertClient(None,"-====================-",bSilentSecureMessages);
		Wrapped_HandleSecureCommand(SuperAdmin,Command,Options);
	}
	else
	{
		Wrapped_HandleSecureCommand(SuperAdmin,Command,Options);
	}
}

// Use only in the case of an emergency!
final function Lockdown()
{
local pawn p;
local int i;
	LogMarked("Server entering lockdown!!!");

	// Eject all players
	for( p = level.pawnlist; p != none; p=p.nextpawn )
	{
		if( p.bIsPlayer )
			P.Destroy();
	}
	
	// Reset passwords
	for( i = 0; i < arraycount(adminpassword); i++ )
		AdminPassword[i] = string(rand(100000));
		
	// Lock server
	GamePassword = string(rand(100000));
}

	// Returns a date & time formatted timestamp.
final function string GetTimestamp()
{
	return Level.Month$"/"$Level.Day$"/"$Level.Year@Level.Hour$":"$Level.Minute$":"$Level.Second;
}

	// Logs with a timestamp.
final function LogMarked(coerce string Data)
{
	Log(GetTimestamp()@Data,LOG_WATERMARK);
}

	// Logs a warning with a timestamp.
final function WarnMarked(coerce string Data)
{
	LogMarked("Warning:"@Data);
}

final function ClearTable()
{
local int i;
	for( i = 0; i < arraycount(AuthorityTable); i++ )
	{
		AuthorityTable[i].Action = "";
		AuthorityTable[i].MinAdminLevel = 0;
	}
	LogMarked("Authority table has been cleared!");
}

final function RehashTable()
{
local int i,j,entries;
local string min;
local int min_index;
local int result;
local S_AuthorityKey Temp;
local string left,right;

	// First compact the table.
	j = 0;
	entries=0;
	for( i = 0; i < (arraycount(authoritytable)); i++ )
	{
		if( AuthorityTable[i].Action != "" )
		{
			if( i != j )
			{
				AuthorityTable[j].Action = AuthorityTable[i].Action;
				AuthorityTable[j].MinAdminLevel = AuthorityTable[i].MinAdminLevel;
				AuthorityTable[j].ActionType = AuthorityTable[i].ActionType;
				AuthorityTable[j].bCovert =  AuthorityTable[i].bCovert;
				
				AuthorityTable[i].Action = "";
				AuthorityTable[i].MinAdminLevel=0;
				AuthorityTable[i].ActionType=0;
				AuthorityTable[i].bCovert = false;
			}
			j++;
			entries++;
		}
	}

	// Reorder the entire fucking authority table. I'm too lazy to write a heap data structure for a more efficient rehash. fuck you.
	for( i = 0; i < (entries-1); i++ )
	{
		min = authoritytable[i].action;
		min_index = i;
		for( j = i+1; j < entries; j++ )
		{
			// Try to find something smaller than min
			left = min$authoritytable[i].actiontype;
			right = authoritytable[j].action$authoritytable[j].actiontype;
			result = CLASS_UTILITY.static.strcmp(left,right);
			if( result == 0 )
			{
				// Ignore this entry entirely. It's useless.
				AuthorityTable[j].action = "";
			}
			else if( result > 0 )
			{
				// MIN > This
				min = AuthorityTable[j].Action;
				min_index = j;
			}
			else
			{
				// MIN < This
				continue;	// keep searching
			}
		}
		
		// Perform swap.
		if( i != min_index )
		{
			Temp = AuthorityTable[i];
			AuthorityTable[i] = AuthorityTable[min_index];
			AuthorityTable[min_index] = Temp;
		}
	}
	
	NumTableEntries   = Entries;
}

/* insertBan 
 * Inserts a new ban in IPv4 dotted IP format with optional rejection message
 */
function bool insertBan( pawn administrator, string IP, optional string newRejectionMessage )
{
local string firstSub,secondSub,thirdSub,fourthSub; // Subnets of the IP
local int index;  // For array traversal of the ban table
local string s;
local playerpawn p;
	if( Administrator == None ) s = "Server";
	else s = Administrator.GetHumanName();
	
  // Validity check: see if this is a valid IPv4 IP address.
  firstSub = CLASS_UTILITY.static.parseNextWord(IP,".");
  secondSub = CLASS_UTILITY.static.parseNextWord(IP,".");
  thirdSub = CLASS_UTILITY.static.parseNextWord(IP,".");
  fourthSub = CLASS_UTILITY.static.parseNextWord(IP,":");
  if( inStr( fourthSub, " " ) > -1 )
    fourthSub = left( fourthSub, inStr( fourthSub, " " ) );
  
  if( firstSub == "" || (firstSub != "*" && !CLASS_UTILITY.static.isNumber(left(firstSub,1) ) ) || 
      secondSub == "" || (secondSub != "*" && !CLASS_UTILITY.static.isNumber(left(secondSub,1) ) ) || 
      thirdSub == "" || (thirdSub != "*" && !CLASS_UTILITY.static.isNumber(left(thirdSub,1) ) ) || 
      fourthSub == "" || (fourthSub != "*" && !CLASS_UTILITY.static.isNumber(left(fourthSub,1) ) ) || 
      CLASS_UTILITY.static.parseNextWord(IP) != "" )
  {
    // invalid
    return false;
  }
  
  // Kick the idiot.
  foreach allactors(class'playerpawn',p)
  {
	if( IPPlugin.static.retrieveIP(level,p) == IP )
		P.Destroy();
  }
  
  // perform actual insertion
  for( index = 0; index < arraycount(banTable); index++ )
  {
    if( banTable[index].banEntryDate == "" )
    {
      // This is empty, insert here
      banTable[index].IP.firstSubnet = firstSub;
      banTable[index].IP.secondSubnet = secondSub;
      banTable[index].IP.thirdSubnet = thirdSub;
      banTable[index].IP.fourthSubnet = fourthSub;
      
      if( playerpawn(Administrator) != none )
      	banTable[index].bannedBy
      	 = s@"("$IPPlugin.static.retrieveIP(level,playerpawn(administrator))$")";
      else
        banTable[index].bannedBy = "(automated)";
        
      banTable[index].banReply = newRejectionMessage;
      
      banTable[index].banEntryDate
       = Level.Month$"/"$Level.Day$"/"$Level.Year$" ("$Level.Hour$":"$Level.Minute$")";
       
      saveConfig();
      
      Log(">>       Alert       <<",LOG_WATERMARK);
      Log(">>  New ban issued  << - Admin "$s@"banned"@firstSub$"."$secondSub$"."$thirdSub$"."$fourthSub@"at slot"@index,LOG_WATERMARK);
      Log(">>       Alert       <<",LOG_WATERMARK);
      return true; // Successful ban
    }
  }
  
  Log("Ban table exhausted! Clean ban table or increase array size!",LOG_WATERMARK);
  return false;
}

/* removeBan
 * Removes a ban entry from a slot
 */
function bool removeBan( pawn administrator, int slot )
{
local string s;
	if( Administrator == None ) s = "Server";
	else s = Administrator.GetHumanName();

  if( slot < 0 || slot >= arrayCount( banTable ) )
  {
    AlertClient(Administrator,"Cannot unban slot "$slot$". Make sure you are attempting to unban the correct slot.",bSilentSecureMessages);
    return false; // invalid
  }
  
  banTable[slot].IP.firstSubnet = "";
  banTable[slot].IP.secondSubnet = "";
  banTable[slot].IP.thirdSubnet = "";
  banTable[slot].IP.fourthSubnet = "";
  
  banTable[slot].bannedBy = "";
  
  banTable[slot].banEntryDate = "";
  
  banTable[slot].banReply = "";
  
  saveConfig();
    
  Log(">>       Alert       <<",LOG_WATERMARK);
  Log(">> Ban entry removed << - "$s@"unbanned slot"@slot,LOG_WATERMARK);
  Log(">>       Alert       <<",LOG_WATERMARK);
  return true; // Successful remove
}

/* isBanned
 * Returns true if Other is banned, false otherwise 
 */
function bool isBanned( pawn Other, out string rejectionMessage )
{
  local int index;  // For array traversal  
  local string IPResult;  // Other's IP in string format'
  local string parse; // Temp result of parsing
  
  if( IPPlugin == none )
  {
    if( other != none )
      warn("Could not assess ban status of"@other$". No valid IPPlugin loaded.");
    else 
      warn("No valid IPPlugin loaded!");
    return false;
  }
  
  if( Other != None && !Other.IsA('Playerpawn') )
	return false;
  IPResult = IPPlugin.Static.retrieveIP(Level,playerpawn(Other)); // get ip
  
  return isIPBanned( IPResult, rejectionMessage );
}

function bool isIPBanned( String IPResult, out String RejectionMessage )
{
local s_IP parsedIP; // Four separated subnets in one structure.. parsed version of IPResult
local int index;

  parsedIP.firstSubnet = CLASS_UTILITY.static.parseNextWord( IPResult, "." );
  parsedIP.secondSubnet = CLASS_UTILITY.static.parseNextWord( IPResult, "." ); // Get second subnet
  parsedIP.thirdSubnet = CLASS_UTILITY.static.parseNextWord( IPResult, "." ); // Get first subnet
  parsedIP.fourthSubnet= CLASS_UTILITY.static.parseNextWord( IPResult, ":" );  // get fourth subnet

  for(index=0;index<arraycount(banTable);index++)
  {
    if( banTable[index].banEntryDate != "" )
    {
      // Check against this ban
      if( (banTable[index].IP.firstSubnet == "*" || int(banTable[index].IP.firstSubnet) == int(parsedIP.firstSubnet)) &&
      (banTable[index].IP.secondSubnet == "*" || int(banTable[index].IP.secondSubnet) == int(parsedIP.secondSubnet)) &&
      (banTable[index].IP.thirdSubnet == "*" || int(banTable[index].IP.thirdSubnet) == int(parsedIP.thirdSubnet)) &&
      (banTable[index].IP.fourthSubnet == "*" || int(banTable[index].IP.fourthSubnet) == int(parsedIP.fourthSubnet)) )
      {
        // This jackass is banned
        if( banTable[index].banReply == "" )
          rejectionMessage = "You are banned from this server. Contact the administrator for further details.";
        else
          rejectionMessage = banTable[index].banReply;

        return true;
      }
    }
  }
  
  return false;
}

function bool IsInServerPackages( string PackageName )
{
local bool bFound;
local string spacks,next;
	// see if it's in serverpackages
	PackageName = caps(PackageName);
	SPacks = caps("\"JGrass\","$mid(consolecommand("Get Engine.GameEngine ServerPackages"),1));
	
	while( SPacks != "" )
	{
		if( instr(SPacks,",") != -1 )
		{
			next = left(SPacks,inStr(SPacks,","));
			sPacks = mid(SPacks,instr(SPacks,",")+1);
		}
		else
		{
			next = left(spacks,len(spacks)-1);
			spacks = "";
		}

		if( next ~= ("\""$PackageName$"\"") )
			bFound = true;
	}
	
	return bFound;
}

defaultproperties
{
				AuthorityTable(0)=(ActionType=1,Action="addbots",MinAdminLevel=3)
				AuthorityTable(1)=(ActionType=1,Action="AddStart",MinAdminLevel=2)
				AuthorityTable(2)=(Action="Admin",MinAdminLevel=1)
				AuthorityTable(3)=(ActionType=2,Action="AllAmmo",MinAdminLevel=3)
				AuthorityTable(4)=(ActionType=2,Action="Amphibious",MinAdminLevel=4)
				AuthorityTable(5)=(Action="ASay",MinAdminLevel=3,bCovert=True)
				AuthorityTable(6)=(Action="BanID",MinAdminLevel=1)
				AuthorityTable(7)=(Action="BanIP",MinAdminLevel=1)
				AuthorityTable(8)=(ActionType=2,Action="duke_giveitems",MinAdminLevel=3)
				AuthorityTable(9)=(ActionType=2,Action="duke_spawnmass",MinAdminLevel=3)
				AuthorityTable(10)=(Action="end",MinAdminLevel=2)
				AuthorityTable(11)=(ActionType=2,Action="Fly",MinAdminLevel=3)
				AuthorityTable(12)=(Action="FreezeID",MinAdminLevel=2)
				AuthorityTable(13)=(Action="GetIDs",MinAdminLevel=4)
				AuthorityTable(14)=(Action="GetIPs",MinAdminLevel=1)
				AuthorityTable(15)=(ActionType=2,Action="Ghost",MinAdminLevel=3)
				AuthorityTable(16)=(Action="GiveAdminID",MinAdminLevel=2)
				AuthorityTable(17)=(ActionType=2,Action="God",MinAdminLevel=3)
				AuthorityTable(18)=(Action="GoToID",MinAdminLevel=3)
				AuthorityTable(19)=(Action="Help",MinAdminLevel=4)
				AuthorityTable(20)=(ActionType=1,Action="Help",MinAdminLevel=4)
				AuthorityTable(21)=(ActionType=2,Action="Help",MinAdminLevel=4)
				AuthorityTable(22)=(Action="KickID",MinAdminLevel=2)
				AuthorityTable(23)=(Action="KillAll",MinAdminLevel=2)
				AuthorityTable(24)=(Action="KillID",MinAdminLevel=2)
				AuthorityTable(25)=(ActionType=1,Action="KNI",MinAdminLevel=3)
				AuthorityTable(26)=(ActionType=1,Action="KNP",MinAdminLevel=3)
				AuthorityTable(27)=(Action="ListBans",MinAdminLevel=2)
				AuthorityTable(28)=(Action="Login",MinAdminLevel=4,bCovert=True)
				AuthorityTable(29)=(Action="Logout",MinAdminLevel=3,bCovert=True)
				AuthorityTable(30)=(Action="MuteID",MinAdminLevel=2)
				AuthorityTable(31)=(ActionType=1,Action="Pause",MinAdminLevel=2)
				AuthorityTable(32)=(Action="PSayID",MinAdminLevel=4,bCovert=True)
				AuthorityTable(33)=(Action="RemoveBan",MinAdminLevel=1)
				AuthorityTable(34)=(ActionType=1,Action="RemoveStarts",MinAdminLevel=2)
				AuthorityTable(35)=(Action="RenameID",MinAdminLevel=2)
				AuthorityTable(36)=(Action="RestartID",MinAdminLevel=2)
				AuthorityTable(37)=(Action="RevokeAdminID",MinAdminLevel=2)
				AuthorityTable(38)=(ActionType=1,Action="Say",MinAdminLevel=4,bCovert=True)
				AuthorityTable(39)=(Action="SetName",MinAdminLevel=4)
				AuthorityTable(40)=(ActionType=1,Action="SloMo",MinAdminLevel=2)
				AuthorityTable(41)=(ActionType=2,Action="Suicide",MinAdminLevel=4)
				AuthorityTable(42)=(ActionType=2,Action="Summon",MinAdminLevel=3)
				AuthorityTable(43)=(Action="SummonID",MinAdminLevel=2)
				AuthorityTable(44)=(Action="SwitchLevel",MinAdminLevel=2)
				AuthorityTable(45)=(ActionType=1,Action="TeamSay",MinAdminLevel=4,bCovert=True)
				AuthorityTable(46)=(Action="UnfreezeID",MinAdminLevel=2)
				AuthorityTable(47)=(Action="UnmuteID",MinAdminLevel=2)
				AuthorityTable(48)=(ActionType=2,Action="ViewClass",MinAdminLevel=4)
				AuthorityTable(49)=(ActionType=2,Action="Walk",MinAdminLevel=3)
				AuthorityTable(50)=(ActionType=3,Action="Zero",MinAdminLevel=1,bCovert=True)
				DefaultAction=(ActionType=-1,Action="Default")
				ActionLoggingType=ALT_All
				MaxRejectionsPerSecond=10
				IPPluginClass="IPPlugin227.IPPlugin227"
				AdminPassword(0)="myumyusoft"
				AdminPassword(1)="trucktastic"
				AdminPassword(2)="zero"
				bAutoBanRJA=True
				MinbAdminLevel=2
				MaxPlayers=6
				MaxMessagesPerSecond=4
				MinAntiSpamOverrideLevel=2
				AntiSpamMuteTime=8
				GeneralSecureMessage="[Secure]"
				MinReceiveSecurityEventsLevel=2
				bBroadcastAuthorizations=True
				bForceSeedNoRespawn=True
}
